import re
import sys
import traceback
from pathlib import Path
from .colored import Attribute, colored

path_pattern = re.compile(r'File "(.*)", line (\d+), in (.+)')
error_pattern = re.compile(r"^[\w.]+: ")

def pretty_print_exc():
  cwd = str(Path.cwd())
  tracelines = traceback.format_exc().splitlines()
  is_external = False
  pretty_lines = [""]

  for line in tracelines:
    matches = path_pattern.match(line.strip())
    if line.startswith("Traceback ("):
      pretty_lines.append(colored(line, attrs=["bold"]))
    elif error_pattern.match(line):
      pretty_lines.append(colored(line, "red"))
    elif not matches:
      pretty_lines.append(colored(line, "dark_grey" if is_external else None))
    else:
      path, line_number, method = matches.groups()
      if path.startswith("<frozen importlib._bootstrap"):
        continue
      is_external = (not path.startswith(".") and not path.startswith(cwd)) or ".venv" in path
      if not is_external:
        path = path.replace(cwd, ".")
      color_attrs: list[Attribute] = ["dark"] * is_external
      pretty_line = (
        f'{colored(path, "cyan", attrs=color_attrs)}'
        f'{colored(":", attrs=color_attrs)}'
        f'{colored(line_number, "yellow", attrs=color_attrs)} '
        f'{colored(method, "green", attrs=color_attrs)}'
        f'{colored(":", attrs=color_attrs)}'
      )
      pretty_lines.append(pretty_line)

  print("\n".join(pretty_lines + [""]), file=sys.stderr)



import io
import os
import sys
from typing import Literal

Color = Literal[
  "black", "grey", "red", "green", "yellow", "blue", "magenta",
  "cyan", "light_grey", "dark_grey", "light_red", "light_green",
  "light_yellow", "light_blue", "light_magenta", "light_cyan", "white",
]

Attribute = Literal["bold", "dark", "underline", "blink", "reverse", "concealed", "strike"]

COLOR_MAP: dict[Color, int] = {
  "black": 30,
  "grey": 30,
  "red": 31,
  "green": 32,
  "yellow": 33,
  "blue": 34,
  "magenta": 35,
  "cyan": 36,
  "light_grey": 37,
  "dark_grey": 90,
  "light_red": 91,
  "light_green": 92,
  "light_yellow": 93,
  "light_blue": 94,
  "light_magenta": 95,
  "light_cyan": 96,
  "white": 97,
}

ATTRIBUTE_MAP: dict[Attribute, int] = {
  "bold": 1,
  "dark": 2,
  "underline": 4,
  "blink": 5,
  "reverse": 7,
  "concealed": 8,
  "strike": 9,
}

def _allow_color() -> bool:
  if "NO_COLOR" in os.environ or os.environ.get("TERM") == "dumb" or not hasattr(sys.stdout, "fileno"):
    return False
  try:
    return os.isatty(sys.stdout.fileno())
  except io.UnsupportedOperation:
    return sys.stdout.isatty()

allow_color = _allow_color()

def colored(text: str, color: Color | None = None, on_color: Color | None = None, attrs: list[Attribute] = []) -> str:
  if not allow_color:
    return text
  if color:
    text = f"\033[{COLOR_MAP[color]}m{text}"
  if on_color:
    text = f"\033[{COLOR_MAP[on_color] + 10}m{text}"
  for attr in attrs:
    text = f"\033[{ATTRIBUTE_MAP[attr]}m{text}"
  return text + "\033[0m"



import shutil
from collections.abc import ItemsView, KeysView, ValuesView
from contextlib import suppress
from .colored import Color, colored

np = None

with suppress(Exception):
  import numpy as np

bracket_colors: list[Color | None] = [
  None,
  "red",
  "blue",
  "magenta",
  "cyan",
  "grey",
]

def get_terminal_width():
  return shutil.get_terminal_size((80, 0))[0]

def render(value, color, indent=0):
  indent_str = "  " * indent
  next_indent_str = indent_str + "  "
  bracket_color = color or bracket_colors[indent % len(bracket_colors)]
  match value:
    case None:
      return colored(str(value), color, attrs=["bold"])
    case _ if value is ...:
      return colored("...", color, attrs=["bold"])
    case int() | float():
      return colored(str(value), color or "yellow")
    case str():
      if indent == 0:
        return colored(f"{value}", color)
      else:
        return colored(f'"{value}"', color or "green")
    case tuple() | set() | frozenset() | list() | KeysView() | ValuesView() | ItemsView():
      if isinstance(value, (frozenset, KeysView, ValuesView, ItemsView)):
        prefix, postfix = f"{type(value).__name__}(", ")"
      else:
        prefix, postfix = "", ""
      opening, closing = "()" if isinstance(value, tuple) \
        else "{}" if isinstance(value, (set, frozenset)) \
        else "[]"
      opening = colored(opening, bracket_color)
      closing = colored(closing, bracket_color)
      comma = "," if isinstance(value, tuple) and len(value) == 1 else ""
      children = [render(val, color, indent + 1) for val in value]
      child_str = ", ".join(children)
      terminal_width = get_terminal_width()
      if len(child_str) <= terminal_width and "\n" not in child_str:
        return f"{prefix}{opening}{child_str}{comma}{closing}{postfix}"
      else:
        children = [next_indent_str + child for child in children]
        child_str = ",\n".join(children)
        return f"{opening}\n{child_str}\n{indent_str}{closing}"
    case dict():
      opening = colored("{", bracket_color)
      closing = colored("}", bracket_color)
      children = [
        f"{next_indent_str}{render(key, color, indent + 1)}: {render(val, color, indent + 1)}"
        for key, val in value.items()
      ]
      child_str = ",\n".join(children)
      child_str = child_str and f"\n{child_str}\n{indent_str}"
      return f"{opening}{child_str}{closing}"
    case _ if np and isinstance(value, np.ndarray):
      array_lines = np.array2string(
        value,
        max_line_width=get_terminal_width(),
        edgeitems=5,
      ).splitlines()
      if len(array_lines) == 1:
        return array_lines[0]
      else:
        array_lines[0] = " " + array_lines[0][1:]
        array_lines[-1] = array_lines[-1][:-1]
        array_lines = [next_indent_str + line for line in array_lines]
        array_str = "\n".join(array_lines)
        return f"ndarray [\n{array_str}\n{indent_str}]"
    case _:
      return str(value)

pyprint = print

def print(*values, color=None, **kwargs):
  pyprint(*(render(value, color) for value in values), **kwargs)



import builtins
import importlib
import os
import signal
import sys
import time
from pathlib import Path
from watchdog.events import PatternMatchingEventHandler
from watchdog.observers import Observer
from . import pretty_print_exc, print, reload_handlers

cwd = Path.cwd().resolve()
module_path = Path(sys.argv[1])
module_path_str = str(module_path.with_suffix("")).replace("/", ".").replace("\\", ".")
is_local_by_module = {}
changed_modules = set()
start_time = 0
historic_local_modname_by_path: dict[Path, str] = {}

builtins.__dict__["print"] = print
sys.path.insert(0, ".")
sys.argv = sys.argv[1:]

def stopwatch():
  global start_time
  duration = round(time.time() - start_time) if start_time else 0
  minutes = duration // 60
  seconds = duration % 60
  minutes_str = f"{minutes}m" if minutes else ""
  duration_str = f"({minutes_str}{seconds}s)" if duration >= 1 else ""
  start_time = 0
  return duration_str

class RestartException(Exception):
  pass

def to_module_path(module):
  return Path(getattr(module, "__file__", None) or "/__").resolve()

def is_local_module(module):
  if module in is_local_by_module:
    return is_local_by_module[module]
  target = to_module_path(module)
  is_local = ".venv" not in target.parts and cwd in target.parents
  is_local_by_module[module] = is_local
  return is_local_module(module)

def get_local_modname_by_path() -> dict[Path, str]:
  result = {
    to_module_path(module): mod_name
    for mod_name, module in list(sys.modules.items())
    if is_local_module(module)
  }
  result[module_path.resolve()] = module_path_str
  return result

def start():
  global start_time
  start_time = time.time()
  try:
    try:
      importlib.import_module(module_path_str)
      historic_local_modname_by_path.clear()
      print(f"⚠️  {module_path} finished. {stopwatch()}")
    except OSError as err:
      if str(err) == "could not get source code":
        start()
      else:
        raise
  except KeyboardInterrupt:
    print(f"\n⚠️  Script interrupted. {stopwatch()}")
  except SystemExit:
    print(f"\n⚠️  Script exited. {stopwatch()}")
  except RestartException:
    pass
  except:
    pretty_print_exc()

def restart(changed_file):
  os.system("cls" if os.name == "nt" else "clear")
  print(f"⚠️  {changed_file.relative_to(cwd)} changed, restarting. {stopwatch()}")
  local_modname_by_path = get_local_modname_by_path()
  historic_local_modname_by_path.update(local_modname_by_path)
  for handler in reload_handlers:
    handler()
  reload_handlers.clear()
  for mod_name in local_modname_by_path.values():
    sys.modules.pop(mod_name, None)
  start()

def receive_signal(signum, stack):
  raise RestartException()

class EventHandler(PatternMatchingEventHandler):
  def on_modified(self, evt):
    src_path = Path(evt.src_path).resolve()
    dest_path = Path(evt.dest_path).resolve() if hasattr(evt, "dest_path") else None
    local_modname_by_path = {**historic_local_modname_by_path, **get_local_modname_by_path()}
    if src_path in local_modname_by_path:
      changed_modules.add(src_path)
    if dest_path in local_modname_by_path:
      changed_modules.add(dest_path)
    if len(changed_modules) and os.name != "nt":
      os.kill(os.getpid(), signal.SIGTERM)

signal.signal(signal.SIGTERM, receive_signal)

observer = Observer()
observer.schedule(EventHandler(patterns=["*.py"]), str(cwd), recursive=True)
observer.start()

def main():
  start()
  while True:
    try:
      mod_path = next(iter(changed_modules), None)
      if mod_path:
        changed_modules.clear()
        restart(mod_path)
      time.sleep(0.05)
    except RestartException:
      pass
    except KeyboardInterrupt:
      break



from typing import Callable
from .colored import colored
from .pretty_print_exc import pretty_print_exc
from .print import print

__all__ = ["colored", "on_reload", "pretty_print_exc", "print"]

reload_handlers: list[Callable] = []

def on_reload(func: Callable):
  reload_handlers.append(func)
